"""Tests for preprocessing converter."""

# coverage erase && coverage run -p -m pytest tests/test_wrapped.py --durations=20 && coverage combine && coverage html && open htmlcov/index.html

import tempfile
import unittest
from pathlib import Path
from typing import ClassVar

from curies import Converter, ReferenceTuple
from curies.preprocessing import (
    BlocklistError,
    PreprocessingBlocklists,
    PreprocessingConverter,
    PreprocessingRewrites,
    PreprocessingRules,
)

EX1_RT = ReferenceTuple("GO", "1234567")
EX1_URI = "http://purl.obolibrary.org/obo/GO_1234567"
EX1_CURIE = "GO:1234567"

DECOY_1_CURIE = "NOPE:NOPE"
DECOY_1_URI = "https://example.org/NOPE/NOPE"


class TestWrapped(unittest.TestCase):
    """Tests for preprocessing converter."""

    rules: ClassVar[PreprocessingRules]
    inner_converter: ClassVar[Converter]
    converter: ClassVar[PreprocessingConverter]
    temporary_directory: ClassVar[tempfile.TemporaryDirectory[str]]
    directory: ClassVar[Path]
    rules_path: ClassVar[Path]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the test case."""
        cls.rules = PreprocessingRules(
            rewrites=PreprocessingRewrites(
                full={
                    "is_a": "rdf:type",
                    "http://creativecommons.org/licenses/by/3.0/": "spdx:CC-BY-3.0",
                },
                prefix={
                    "OMIM:PS": "omim.ps:",
                    "omim:PS": "omim.ps:",
                },
                resource_prefix={
                    "clo": {
                        "j": "NCIT:",
                    },
                },
                resource_full={
                    "clo": {
                        "nopeforever": "NCIT:5678",
                    },
                },
            ),
            blocklists=PreprocessingBlocklists(
                full=["rdf:NOPE"],
                resource_prefix={
                    "chebi": [
                        "pubmed:"
                    ]  # this means that only throw away pubmed references in ChEBI
                },
                resource_full={
                    "chebi": ["omim:1356"],  # in case we just hate this CURIE/URI/string
                },
            ),
        )
        cls.inner_converter = Converter.from_prefix_map(
            {
                "GO": "http://purl.obolibrary.org/obo/GO_",
                "NCIT": "http://purl.obolibrary.org/obo/NCIT_",
                "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "omim": "https://omim.org/MIM:",
                "omim.ps": "https://omim.org/phenotypicSeries/PS",
                "pubmed": "http://rdf.ncbi.nlm.nih.gov/pubchem/reference/",
                "spdx": "https://spdx.org/licenses/",
            }
        )

        cls.temporary_directory = tempfile.TemporaryDirectory()
        cls.directory = Path(cls.temporary_directory.name)

        # write the rules just so we can test loading from a file
        cls.rules_path = cls.directory.joinpath("rules.json")
        cls.rules_path.write_text(cls.rules.model_dump_json(indent=2, exclude_unset=True))

        cls.converter = PreprocessingConverter.from_converter(
            converter=cls.inner_converter, rules=cls.rules_path
        )

    @classmethod
    def tearDownClass(cls) -> None:
        """Clean up the class."""
        cls.temporary_directory.cleanup()

    def test_lint(self) -> None:
        """Run the linting code, just to make sure it works."""
        self.rules.lint_file(self.rules_path)

    def test_unprocessed(self) -> None:
        """Tests that should be the same for both converters."""
        for converter in (self.inner_converter, self.converter):
            self.assertEqual(EX1_CURIE, converter.compress(EX1_URI))
            self.assertEqual(EX1_URI, converter.expand(EX1_CURIE))
            self.assertEqual(EX1_RT, converter.parse(EX1_URI))
            self.assertEqual(EX1_RT, converter.parse(EX1_CURIE))
            self.assertEqual(EX1_RT, converter.parse_uri(EX1_URI))
            self.assertEqual(EX1_RT, converter.parse_curie(EX1_CURIE))

            self.assertIsNone(converter.compress(DECOY_1_URI))
            self.assertIsNone(converter.expand(DECOY_1_CURIE))
            self.assertIsNone(converter.parse_curie(DECOY_1_CURIE))
            self.assertIsNone(converter.parse_uri(DECOY_1_URI, return_none=True))
            self.assertIsNone(converter.parse(DECOY_1_URI))
            self.assertIsNone(converter.parse(DECOY_1_CURIE))

    def test_global_full_rewrite(self) -> None:
        """Test global full string rewrite."""
        self.assertEqual(ReferenceTuple("rdf", "type"), self.converter.parse("is_a"))
        self.assertEqual(ReferenceTuple("rdf", "type"), self.converter.parse_curie("is_a"))
        self.assertEqual(
            ReferenceTuple("spdx", "CC-BY-3.0"),
            self.converter.parse_uri("http://creativecommons.org/licenses/by/3.0/"),
        )

    def test_global_prefix_rewrite(self) -> None:
        """Test global prefix rewrite."""
        self.assertEqual(
            ReferenceTuple("omim.ps", "1234"), self.converter.parse_curie("OMIM:PS1234")
        )

        # test when there's a related rewrite rule, but not used
        self.assertEqual(ReferenceTuple("omim", "1234"), self.converter.parse_curie("omim:1234"))

    def test_resource_full_rewrite(self) -> None:
        """Test global full string rewrite."""
        self.assertIsNone(self.converter.parse("nopeforever"))
        with self.assertRaises(ValueError):
            self.converter.parse_curie("nopeforever")
        self.assertEqual(
            ReferenceTuple("NCIT", "5678"),
            self.converter.parse_curie("nopeforever", context="clo"),
        )

    def test_resource_prefix_rewrite(self) -> None:
        """Test resource-specific prefix rewrite."""
        self.assertEqual(
            ReferenceTuple("NCIT", "1234"),
            self.converter.parse("j1234", context="clo"),
        )

        # when we have rewrite rules for that ontology, but none apply
        self.assertIsNone(self.converter.parse("xyz", context="clo"))

        with self.assertRaises(ValueError):
            self.assertIsNone(self.converter.parse_curie("j1234"))
        with self.assertRaises(ValueError):
            self.assertIsNone(self.converter.parse_curie("j1234", context="chebi"))

    def test_resource_specific_blocklist(self) -> None:
        """Test resource-specific blocklist."""
        self.assertEqual(
            ReferenceTuple("pubmed", "1234"),
            self.converter.parse_curie("pubmed:1234"),
        )
        self.assertEqual(
            ReferenceTuple("pubmed", "1234"),
            self.converter.parse_curie("pubmed:1234", context="doid"),
        )
        with self.assertRaises(BlocklistError):
            self.converter.parse_curie("pubmed:1234", context="chebi")
        self.assertIsNone(
            self.converter.parse_curie("pubmed:1234", context="chebi", block_action="pass")
        )

        self.converter.parse_curie("omim:1234", context="chebi")
        # normally, OMIM works, but we configured a specific one for the blocklist
        with self.assertRaises(BlocklistError):
            self.converter.parse_curie("omim:1356", context="chebi")
        self.assertIsNone(
            self.converter.parse_curie("omim:1356", context="chebi", block_action="pass")
        )

    def test_global_blocklist(self) -> None:
        """Test global blocklist."""
        with self.assertRaises(BlocklistError):
            self.converter.parse("rdf:NOPE")
        self.assertIsNone(self.converter.parse("rdf:NOPE", block_action="pass"))

        with self.assertRaises(BlocklistError):
            self.converter.parse_curie("rdf:NOPE")
        self.assertIsNone(self.converter.parse_curie("rdf:NOPE", block_action="pass"))

        with self.assertRaises(BlocklistError):
            self.converter.parse_uri("rdf:NOPE")
        self.assertIsNone(self.converter.parse_uri("rdf:NOPE", block_action="pass"))



"""Validation based on W3C standards.

The Worldwide Web Consortium (W3C) provides standards for
`prefixes <https://www.w3.org/TR/1999/REC-xml-names-19990114/#NT-NCName>`_ (i.e., ``NCName``),
`CURIEs <https://www.w3.org/TR/2010/NOTE-curie-20101216/>`_, and
`IRIs <https://www.ietf.org/rfc/rfc3987.txt>`_, but they are
highly obfuscated and spread across many documents.
This module attempts to operationalize these standards, along with best practices
of documentation and testing.

.. seealso::

    Some other work towards operationalizing these standards:

    - https://github.com/linkml/linkml-runtime/blob/main/linkml_runtime/utils/uri_validator.py
    - https://github.com/dgerber/rfc3987/blob/gh-archived/rfc3987.py

"""

import re

__all__ = [
    "CURIE_PATTERN",
    "LOCAL_UNIQUE_IDENTIFIER_PATTERN",
    "NCNAME_PATTERN",
    "is_w3c_curie",
    "is_w3c_prefix",
]

NCNAME_PATTERN = r"[A-Za-z_][A-Za-z0-9\.\-_]*"
"""A regex for prefixes, from https://www.w3.org/TR/1999/REC-xml-names-19990114/#NT-NCName.

.. code-block::

    prefix := NCName
    NCName := (Letter | '_') (NCNameChar)*
    NCNameChar	::=	Letter | Digit | '.' | '-' | '_'
"""

NCNAME_RE = re.compile(f"^{NCNAME_PATTERN}$")

LOCAL_UNIQUE_IDENTIFIER_PATTERN = r"(/[^\s/][^\s]*|[^\s/][^\s]*|[^\s]?)"
"""A regex for local unique identifiers in CURIEs, based on https://www.ietf.org/rfc/rfc3987.txt

This pattern was adapted from https://gist.github.com/niklasl/2506955, which sort of
implements RFC3987,
"""

LOCAL_UNIQUE_IDENTIFIER_RE = re.compile(LOCAL_UNIQUE_IDENTIFIER_PATTERN)

CURIE_PATTERN = rf"^({NCNAME_PATTERN}?:)?{LOCAL_UNIQUE_IDENTIFIER_PATTERN}$"
"""A regex for CURIEs, based on https://www.w3.org/TR/2010/NOTE-curie-20101216.

.. code-block::

    curie       :=   [ [ prefix ] ':' ] reference
    prefix      :=   NCName
    reference   :=   irelative-ref (as defined in `IRI <https://www.ietf.org/rfc/rfc3987.txt>`_)

`irelative-ref` is defined/documented in :data:`curies.w3c.LOCAL_UNIQUE_IDENTIFIER_PATTERN`.
"""

CURIE_RE = re.compile(CURIE_PATTERN)


def is_w3c_prefix(prefix: str) -> bool:
    """Return if the string is a valid prefix under the W3C specification.

    :param prefix: A string
    :return: If the string is a valid prefix under the W3C specification.

    Validation is implemented as a regular expression match against
    :data:`curies.w3c.NCNAME_PATTERN`, as defined by the W3C
    `here <https://www.w3.org/TR/1999/REC-xml-names-19990114/#NT-NCName>`_.

    Examples
    --------
    Strings containig numbers, letters, and underscores are valid prefixes.

    >>> is_w3c_prefix("GO")
    True

    The W3C specification states that the prefix '_' is reserved for use
    by languages that support RDF. For this reason, the prefix '_' SHOULD
    be avoided by authors.

    >>> is_w3c_prefix("_")
    True

    Strings starting with a number are not
    valid prefixes.

    >>> is_w3c_prefix("3dmet")

    Strings containing a colon or other
    characters are invalid

    >>> is_w3c_prefix("GO:")
    False
    """
    return bool(NCNAME_RE.match(prefix))


def _is_w3c_luid(luid: str) -> bool:
    return bool(LOCAL_UNIQUE_IDENTIFIER_RE.match(luid))


def is_w3c_curie(curie: str) -> bool:
    """Return if the string is a valid CURIE under the W3C specification.

    :param curie: A string to check if it is a valid CURIE under the W3C specification.
    :return: True if the string is a valid CURIE under the W3C specification.

    .. warning::

        This is slightly different from the :meth:`curies.Converter.is_curie` function,
        which checks if a given CURIE is valid under the extended prefix map contained
        within the converter.

        Further, the base converter is slightly more lenient than the W3C specification
        by default to allow for the inclusion of CURIEs, e.g., for SMILES strings like
        ``smiles:CC(=O)NC([H])(C)C(=O)O``. These are useful, but not technically valid
        due to their inclusion of brackets.

    Examples
    --------
    If no prefix is given, the host language chooses how to assign a default
    prefix.

    >>> is_w3c_curie(":test")
    True

    From the specification, regarding using an underscore as the prefix

        The CURIE prefix '_' is reserved for use by languages that support RDF.
        For this reason, the prefix '_' SHOULD be avoided by authors.

    >>> is_w3c_curie("_:test")
    True

    This is invalid because a CURIE prefix isn't allowed to start with
    a number. It has to start with either a letter, or an underscore.

    >>> is_w3c_curie("4cdn:test")
    False

    Empty strings are explicitly noted as being invalid.

    >>> is_w3c_curie("")
    False
    """
    if "[" in curie or "]" in curie:
        return False

    # empty curie is invalid (for now)
    if not curie.strip():
        return False

    # if there's no colon, then validate the whole thing against the LUID pattern.
    # this is because
    prefix, sep, identifier = curie.partition(":")
    if not sep:
        return _is_w3c_luid(curie)

    # it's okay for there to be no prefix in a CURIE, even though
    # the NCName definition is not itself allowed to be empty
    if not prefix:
        return _is_w3c_luid(identifier)

    return is_w3c_prefix(prefix) and _is_w3c_luid(identifier)



"""Test types."""

import unittest

from pydantic import BaseModel, ValidationError

from curies import Converter, Prefix, PrefixMap, Reference


class WrappedPrefix(BaseModel):
    """A model wrapping a prefix."""

    prefix: Prefix


class WrappedPrefixMap(BaseModel):
    """A model wrapping a prefix map."""

    prefix_map: PrefixMap


class WrappedCURIE(BaseModel):
    """A model wrapping a reference."""

    reference: Reference


converter = Converter.from_extended_prefix_map(
    [
        {
            "prefix": "CHEBI",
            "prefix_synonyms": ["chebi"],
            "uri_prefix": "http://purl.obolibrary.org/obo/CHEBI_",
            "uri_prefix_synonyms": [
                "https://identifiers.org/chebi:",
            ],
        },
    ]
)


class TestTypes(unittest.TestCase):
    """Test types."""

    def test_prefix(self) -> None:
        """Test instantiating prefixes."""
        model_1 = WrappedPrefix.model_validate({"prefix": "hello"})
        self.assertEqual("hello", model_1.prefix)

        model_2 = WrappedPrefix.model_validate({"prefix": "CHEBI"})
        self.assertEqual("CHEBI", model_2.prefix)

        # Test that a synonym gets standardized properly
        model_3 = WrappedPrefix.model_validate({"prefix": "CHEBI"}, context=converter)
        self.assertEqual("CHEBI", model_3.prefix)

        # Test that a canonical prefix is passed through
        model_4 = WrappedPrefix.model_validate({"prefix": "chebi"}, context=converter)
        self.assertEqual("CHEBI", model_4.prefix)

        # Test that a synonym gets standardized properly, when passing context in a dict
        model_5 = WrappedPrefix.model_validate(
            {"prefix": "CHEBI"}, context={"converter": converter}
        )
        self.assertEqual("CHEBI", model_5.prefix)

        # Test that a canonical prefix is passed through, when passing context in a dict
        model_6 = WrappedPrefix.model_validate(
            {"prefix": "chebi"}, context={"converter": converter}
        )
        self.assertEqual("CHEBI", model_6.prefix)

        # Test an invalid prefix raises an error, when passing a converter directly
        with self.assertRaises(ValidationError):
            WrappedPrefix.model_validate({"prefix": "nope"}, context=converter)

        # Test an invalid prefix raises an error, when passing a converter in a dict
        with self.assertRaises(ValidationError):
            WrappedPrefix.model_validate({"prefix": "nope"}, context={"converter": converter})

    def test_prefix_root_model(self) -> None:
        """Test the root model."""
        dd = {
            "": "http://example.org",
            "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
        }
        prefix_map = PrefixMap.model_validate(dd)
        self.assertEqual(dd, prefix_map.root)

        prefix_map = PrefixMap.model_validate(
            {
                "chebi": "http://purl.obolibrary.org/obo/CHEBI_",
            },
            context=converter,
        )
        self.assertEqual(
            {
                "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
            },
            prefix_map.root,
        )

        with self.assertRaises(ValidationError):
            PrefixMap.model_validate(
                {
                    "NOPE": "http://purl.obolibrary.org/obo/CHEBI_",
                },
                context=converter,
            )

    def test_prefix_map_wrapped(self) -> None:
        """Test a wrapped prefix map."""
        dd = {
            "prefix_map": {
                "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
            }
        }
        wpm = WrappedPrefixMap.model_validate(dd)
        self.assertIn("CHEBI", wpm.prefix_map.root)

    def test_curie(self) -> None:
        """Test a wrapped CURIE."""
        wpm = WrappedCURIE.model_validate(
            {
                "reference": "CHEBI:1234",
            }
        )
        self.assertIn("CHEBI", wpm.reference.prefix)
        self.assertIn("1234", wpm.reference.identifier)
        self.assertIn("CHEBI:1234", wpm.reference.curie)

        with self.assertRaises(ValidationError):
            WrappedCURIE.model_validate({"reference": "NOPENOPENOPE"})

        dd = {"reference": "CHEBI:1234"}
        wpm = WrappedCURIE.model_validate(dd, context=converter)
        self.assertIn("CHEBI", wpm.reference.prefix)
        self.assertIn("1234", wpm.reference.identifier)
        self.assertIn("CHEBI:1234", wpm.reference.curie)

        with self.assertRaises(ValidationError):
            WrappedCURIE.model_validate({"reference": "MONDO:1234"}, context=converter)



"""Trivial version test."""

from __future__ import annotations

import json
import tempfile
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import pandas as pd
import rdflib
from pydantic import ValidationError

import curies
from curies.api import (
    CompressionError,
    Converter,
    CURIEStandardizationError,
    DuplicatePrefixes,
    DuplicateURIPrefixes,
    ExpansionError,
    NamableReference,
    NamedReference,
    NoCURIEDelimiterError,
    PrefixStandardizationError,
    Record,
    Records,
    Reference,
    ReferenceTuple,
    URIStandardizationError,
    chain,
    upgrade_prefix_map,
)
from curies.sources import (
    BIOREGISTRY_CONTEXTS,
    get_bioregistry_converter,
    get_go_converter,
    get_monarch_converter,
    get_obo_converter,
)
from curies.version import get_version
from tests.constants import SLOW

CHEBI_URI_PREFIX = "http://purl.obolibrary.org/obo/CHEBI_"
GO_URI_PREFIX = "http://purl.obolibrary.org/obo/GO_"


class TestStruct(unittest.TestCase):
    """Test the data structures."""

    def test_not_curie(self) -> None:
        """Test a malformed CURIE."""
        with self.assertRaises(NoCURIEDelimiterError) as e:
            Reference.from_curie("not a curie")
        self.assertIn("does not appear to be a CURIE", str(e.exception))

    def test_default_prefix(self) -> None:
        """Test a default (empty) prefix."""
        ref = Reference.from_curie(":something")
        self.assertEqual("", ref.prefix)
        self.assertEqual("something", ref.identifier)

    def test_default_identifier(self) -> None:
        """Test a default (empty) identifier."""
        ref = Reference.from_curie("p1:")
        self.assertEqual("p1", ref.prefix)
        self.assertEqual("", ref.identifier)

    def test_multiple_delimiters(self) -> None:
        """Test a default (empty) identifier."""
        ref = Reference.from_curie("a1:b2:c3")
        self.assertEqual("a1", ref.prefix)
        self.assertEqual("b2:c3", ref.identifier)

    def test_records(self) -> None:
        """Test a list of records."""
        records = Records.model_validate([{"prefix": "chebi", "uri_prefix": CHEBI_URI_PREFIX}])
        converter = Converter(records=records)
        self.assertEqual({"chebi"}, converter.get_prefixes())

    def test_sort(self) -> None:
        """Test sorting."""
        start = [
            Reference.from_curie("def:1234"),
            Reference.from_curie("abc:1234"),
            Reference.from_curie("abc:1235"),
        ]
        expected = [
            Reference.from_curie("abc:1234"),
            Reference.from_curie("abc:1235"),
            Reference.from_curie("def:1234"),
        ]
        self.assertEqual(expected, sorted(start))

    def test_set_membership(self) -> None:
        """Test membership in sets."""
        collection = {
            Reference.from_curie("def:1234"),
            Reference.from_curie("abc:1234"),
            Reference.from_curie("abc:1235"),
        }
        self.assertIn(Reference.from_curie("def:1234"), collection)
        self.assertNotIn(Reference.from_curie("xyz:1234"), collection)
        self.assertNotIn(Reference.from_curie(":1234"), collection)
        self.assertNotIn(Reference.from_curie("abc:"), collection)

    def test_named_set_membership(self) -> None:
        """Test membership in sets of named references."""
        references = {
            NamedReference.from_curie("a:1", "name1"),
            NamedReference.from_curie("a:2", "name2"),
        }
        self.assertIn(Reference.from_curie("a:1"), references)
        self.assertIn(NamableReference.from_curie("a:1"), references)
        self.assertIn(NamedReference.from_curie("a:1", "name1"), references)
        self.assertIn(NamableReference.from_curie("a:1", "name1"), references)
        # the following is a weird case, but shows how this works
        self.assertIn(NamedReference.from_curie("a:1", "name2"), references)

        references_2 = {
            Reference.from_curie("a:1"),
            Reference.from_curie("a:2"),
        }
        self.assertIn(Reference.from_curie("a:1"), references_2)
        self.assertIn(NamableReference.from_curie("a:1", "name1"), references_2)
        self.assertIn(NamedReference.from_curie("a:1", "name1"), references_2)

    def test_tuple(self) -> None:
        """Test reference tuples."""
        t = ReferenceTuple.from_curie("a:1")
        self.assertEqual(Reference(prefix="a", identifier="1"), t.to_pydantic())

    def test_reference_constructor(self) -> None:
        """Test constructing a reference."""
        r1 = Reference(prefix="a", identifier="1")
        r2 = NamableReference(prefix="a", identifier="2")
        r3 = NamableReference(prefix="a", identifier="3", name="item 3")
        r4 = NamedReference(prefix="a", identifier="4", name="item 4")

        self.assertEqual(Reference(prefix="a", identifier="1"), Reference.from_reference(r1))
        self.assertEqual(Reference(prefix="a", identifier="2"), Reference.from_reference(r2))
        self.assertEqual(Reference(prefix="a", identifier="3"), Reference.from_reference(r3))
        self.assertEqual(Reference(prefix="a", identifier="4"), Reference.from_reference(r4))

        self.assertEqual(
            NamableReference(prefix="a", identifier="1", name=None),
            NamableReference.from_reference(r1),
        )
        self.assertEqual(
            NamableReference(prefix="a", identifier="2", name=None),
            NamableReference.from_reference(r2),
        )
        self.assertEqual(
            NamableReference(prefix="a", identifier="3", name="item 3"),
            NamableReference.from_reference(r3),
        )
        self.assertEqual(
            NamableReference(prefix="a", identifier="4", name="item 4"),
            NamableReference.from_reference(r4),
        )

        with self.assertRaises(TypeError):
            NamedReference.from_reference(r1)
        with self.assertRaises(ValidationError):
            NamedReference.from_reference(r2)
        self.assertEqual(
            NamedReference(prefix="a", identifier="3", name="item 3"),
            NamedReference.from_reference(r3),
        )
        self.assertEqual(
            NamedReference(prefix="a", identifier="4", name="item 4"),
            NamedReference.from_reference(r4),
        )


class TestAddRecord(unittest.TestCase):
    """Test adding records."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.prefix = "CHEBI"
        self.uri_prefix = CHEBI_URI_PREFIX
        self.prefix_synonym = "p"
        self.uri_prefix_synonym = "u"
        self.converter = Converter.from_extended_prefix_map(
            [
                {
                    "prefix": self.prefix,
                    "prefix_synonyms": [self.prefix_synonym],
                    "uri_prefix": self.uri_prefix,
                    "uri_prefix_synonyms": [self.uri_prefix_synonym],
                },
            ]
        )

    def test_duplicate_failure(self) -> None:
        """Test failure caused by double matching."""
        self.converter.add_prefix("GO", GO_URI_PREFIX)
        with self.assertRaises(ValueError):
            self.converter.add_record(Record(prefix="GO", uri_prefix=CHEBI_URI_PREFIX))

    def test_get_prefix_synonyms(self) -> None:
        """Test getting prefix synonyms."""
        self.assertEqual({self.prefix}, self.converter.get_prefixes())
        self.assertEqual({self.prefix}, self.converter.get_prefixes(include_synonyms=False))
        self.assertEqual(
            {self.prefix, self.prefix_synonym},
            self.converter.get_prefixes(include_synonyms=True),
        )

    def test_get_uri_prefix_synonyms(self) -> None:
        """Test getting URI prefix synonyms."""
        self.assertEqual({self.uri_prefix}, self.converter.get_uri_prefixes())
        self.assertEqual({self.uri_prefix}, self.converter.get_uri_prefixes(include_synonyms=False))
        self.assertEqual(
            {self.uri_prefix, self.uri_prefix_synonym},
            self.converter.get_uri_prefixes(include_synonyms=True),
        )

    def test_extend_on_prefix_match(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        for record in [
            Record(
                prefix="CHEBI",
                prefix_synonyms=[s1],
                uri_prefix=s2,
                uri_prefix_synonyms=[s3],
            ),
            Record(
                prefix=s1,
                prefix_synonyms=["CHEBI"],
                uri_prefix=s2,
                uri_prefix_synonyms=[s3],
            ),
        ]:
            with self.assertRaises(ValueError):
                self.converter.add_record(record, merge=False)
            self.converter.add_record(record, merge=True)
            self.assertEqual(1, len(self.converter.records))
            record = self.converter.records[0]
            self.assertEqual("CHEBI", record.prefix)
            self.assertEqual({s1, self.prefix_synonym}, set(record.prefix_synonyms))
            self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
            self.assertEqual({s2, s3, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))

    def test_extend_on_uri_prefix_match(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        for record in [
            Record(
                prefix=s1,
                prefix_synonyms=[s3],
                uri_prefix=s2,
                uri_prefix_synonyms=[CHEBI_URI_PREFIX],
            ),
            Record(
                prefix=s1,
                prefix_synonyms=[s3],
                uri_prefix=CHEBI_URI_PREFIX,
                uri_prefix_synonyms=[s2],
            ),
        ]:
            with self.assertRaises(ValueError):
                self.converter.add_record(record, merge=False)
            self.converter.add_record(record, merge=True)
            self.assertEqual(1, len(self.converter.records))
            record = self.converter.records[0]
            self.assertEqual("CHEBI", record.prefix)
            self.assertEqual({s1, s3, self.prefix_synonym}, set(record.prefix_synonyms))
            self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
            self.assertEqual({s2, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))

    def test_extend_on_prefix_synonym_match(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        for record in [
            Record(
                prefix=self.prefix_synonym,
                prefix_synonyms=[s1],
                uri_prefix=s2,
                uri_prefix_synonyms=[s3],
            ),
            Record(
                prefix=s1,
                prefix_synonyms=[self.prefix_synonym],
                uri_prefix=s2,
                uri_prefix_synonyms=[s3],
            ),
        ]:
            self.converter.add_record(record, merge=True)
            self.assertEqual(1, len(self.converter.records))
            record = self.converter.records[0]
            self.assertEqual("CHEBI", record.prefix)
            self.assertEqual({s1, self.prefix_synonym}, set(record.prefix_synonyms))
            self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
            self.assertEqual({s2, s3, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))

    def test_extend_on_uri_prefix_synonym_match(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        for record in [
            Record(
                prefix=s1,
                prefix_synonyms=[s2],
                uri_prefix=self.uri_prefix_synonym,
                uri_prefix_synonyms=[s3],
            ),
            Record(
                prefix=s1,
                prefix_synonyms=[s2],
                uri_prefix=s3,
                uri_prefix_synonyms=[self.uri_prefix_synonym],
            ),
        ]:
            self.converter.add_record(record, merge=True)
            self.assertEqual(1, len(self.converter.records))
            record = self.converter.records[0]
            self.assertEqual("CHEBI", record.prefix)
            self.assertEqual({s1, s2, self.prefix_synonym}, set(record.prefix_synonyms))
            self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
            self.assertEqual({s3, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))

    def test_extend_on_prefix_match_ci(self) -> None:
        """Test adding a new prefix in merge mode."""
        s1, s2, s3 = "s1", "s2", "s3"
        record = Record(
            prefix="chebi", prefix_synonyms=[s1], uri_prefix=s2, uri_prefix_synonyms=[s3]
        )
        self.converter.add_record(record, case_sensitive=False, merge=True)
        self.assertEqual(1, len(self.converter.records))
        record = self.converter.records[0]
        self.assertEqual("CHEBI", record.prefix)
        self.assertEqual({"chebi", s1, self.prefix_synonym}, set(record.prefix_synonyms))
        self.assertEqual(CHEBI_URI_PREFIX, record.uri_prefix)
        self.assertEqual({s2, s3, self.uri_prefix_synonym}, set(record.uri_prefix_synonyms))


class TestConverter(unittest.TestCase):
    """Test the converter class."""

    def setUp(self) -> None:
        """Set up the converter test case."""
        self.simple_obo_prefix_map = {
            "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
            "MONDO": "http://purl.obolibrary.org/obo/MONDO_",
            "GO": "http://purl.obolibrary.org/obo/GO_",
            "OBO": "http://purl.obolibrary.org/obo/",
        }
        self.converter = Converter.from_prefix_map(self.simple_obo_prefix_map)

    def test_reference_tuple(self) -> None:
        """Test the reference tuple data type."""
        t = ReferenceTuple("chebi", "1234")
        self.assertEqual("chebi:1234", t.curie)
        self.assertEqual(t, ReferenceTuple.from_curie("chebi:1234"))

    def test_reference_pydantic(self) -> None:
        """Test the reference Pydantic model."""
        t = Reference(prefix="chebi", identifier="1234")
        self.assertEqual("chebi:1234", t.curie)
        self.assertEqual(t, Reference.from_curie("chebi:1234"))

    def test_invalid_record(self) -> None:
        """Test throwing an error for invalid records."""
        with self.assertRaises(ValueError):
            Record(
                prefix="chebi",
                uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                prefix_synonyms=["chebi"],
            )
        with self.assertRaises(ValueError):
            Record(
                prefix="chebi",
                uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                uri_prefix_synonyms=["http://purl.obolibrary.org/obo/CHEBI_"],
            )

    def test_invalid_records(self) -> None:
        """Test throwing an error for duplicated URI prefixes."""
        with self.assertRaises(DuplicateURIPrefixes) as e:
            curies.load_prefix_map(
                {
                    "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
                    "nope": "http://purl.obolibrary.org/obo/CHEBI_",
                }
            )
        self.assertIsInstance(str(e.exception), str)
        with self.assertRaises(DuplicatePrefixes) as e2:
            Converter(
                [
                    Record(prefix="chebi", uri_prefix="https://bioregistry.io/chebi:"),
                    Record(prefix="chebi", uri_prefix="http://purl.obolibrary.org/obo/CHEBI_"),
                ],
            )
        self.assertIsInstance(str(e2.exception), str)

        # No failure
        Converter.from_prefix_map(
            {
                "CHEBI": "http://purl.obolibrary.org/obo/CHEBI_",
                "nope": "http://purl.obolibrary.org/obo/CHEBI_",
            },
            strict=False,
        )

    def test_subset(self) -> None:
        """Test subsetting a converter."""
        new_converter = self.converter.get_subconverter(["CHEBI"])
        self.assertEqual(1, len(new_converter.records))
        self.assertEqual({"CHEBI"}, new_converter.get_prefixes())
        self.assertEqual(
            {"http://purl.obolibrary.org/obo/CHEBI_"}, new_converter.get_uri_prefixes()
        )
        self.assertEqual({"CHEBI"}, set(new_converter.bimap))
        self.assertEqual({"CHEBI"}, set(new_converter.prefix_map))
        self.assertEqual(
            {"http://purl.obolibrary.org/obo/CHEBI_"}, set(new_converter.reverse_prefix_map)
        )

    def test_empty_subset(self) -> None:
        """Test subsetting a converter and getting an empty one back."""
        new_converter_2 = self.converter.get_subconverter(["NOPE"])
        self.assertEqual(0, len(new_converter_2.records))

    def test_predicates(self) -> None:
        """Add tests for predicates."""
        self.assertFalse(self.converter.is_uri(""))
        self.assertFalse(self.converter.is_uri("nope"))
        self.assertFalse(self.converter.is_curie(""))
        self.assertFalse(self.converter.is_curie("nope"))
        self.assertFalse(self.converter.is_curie(":nope"))
        self.assertFalse(self.converter.is_curie("nope:"))

    def test_convert(self) -> None:
        """Test compression."""
        self.assertEqual({"CHEBI", "MONDO", "GO", "OBO"}, self.converter.get_prefixes())
        self.assertEqual(
            {
                "http://purl.obolibrary.org/obo/CHEBI_",
                "http://purl.obolibrary.org/obo/MONDO_",
                "http://purl.obolibrary.org/obo/GO_",
                "http://purl.obolibrary.org/obo/",
            },
            self.converter.get_uri_prefixes(),
        )
        self._assert_convert(self.converter)

    def _assert_convert(self, converter: Converter) -> None:
        self.assertIn("GO", converter.prefix_map)
        self.assertIn("GO", converter.bimap)
        self.assertIn("GO", converter.reverse_bimap.values())
        self.assertIn("http://purl.obolibrary.org/obo/GO_", converter.reverse_prefix_map)
        self.assertIn("http://purl.obolibrary.org/obo/GO_", converter.reverse_bimap)
        self.assertIn("http://purl.obolibrary.org/obo/GO_", converter.trie)
        self.assertIn("http://purl.obolibrary.org/obo/GO_", converter.bimap.values())
        for curie, uri in [
            ("CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
            ("OBO:unnamespaced", "http://purl.obolibrary.org/obo/unnamespaced"),
        ]:
            self.assertTrue(converter.is_uri(uri))
            self.assertTrue(converter.is_curie(curie))
            self.assertFalse(converter.is_curie(uri))
            self.assertFalse(converter.is_uri(curie))
            self.assertEqual(curie, converter.compress(uri))
            self.assertEqual(curie, converter.compress_strict(uri))
            self.assertEqual(uri, converter.expand(curie))
            self.assertEqual(uri, converter.expand_strict(curie))

        self.assertIsNone(converter.compress("http://example.org/missing:00000"))
        self.assertEqual(
            "http://example.org/missing:00000",
            converter.compress("http://example.org/missing:00000", passthrough=True),
        )
        with self.assertRaises(CompressionError):
            converter.compress_strict("http://example.org/missing:00000")

        self.assertIsNone(converter.expand("missing:00000"))
        self.assertEqual("missing:00000", converter.expand("missing:00000", passthrough=True))
        with self.assertRaises(ExpansionError):
            converter.expand_strict("missing:00000")

        self.assertLess(0, len(converter.records), msg="converter has no records")
        self.assertIsNone(converter.get_record("nope"))
        self.assertIsNone(converter.get_record("go"), msg="synonym lookup is not allowed here")
        record = converter.get_record("GO", strict=True)
        self.assertIsNotNone(record, msg=f"records: {[r.prefix for r in converter.records]}")
        self.assertIsInstance(record, Record)
        self.assertEqual("GO", record.prefix)

    @SLOW
    def test_bioregistry(self) -> None:
        """Test loading a remote JSON-LD context."""
        for web in [True, False]:
            bioregistry_converter = get_bioregistry_converter(web=web)
            self.assert_bioregistry_converter(bioregistry_converter)

        c = Converter.from_reverse_prefix_map(f"{BIOREGISTRY_CONTEXTS}/bioregistry.rpm.json")
        self.assertIn("chebi", c.prefix_map)
        self.assertNotIn("CHEBI", c.prefix_map)

    def test_jsonld(self) -> None:
        """Test parsing JSON-LD context."""
        context = {
            "@context": {
                "@version": "1.0.0",  # should skip this
                "": "",  # should skip this
                "hello": "https://example.org/hello:",
                "CHEBI": {
                    "@prefix": True,
                    "@id": "http://purl.obolibrary.org/CHEBI_",
                },
                "nope": {
                    "nope": "nope",
                },
            },
        }
        converter = Converter.from_jsonld(context)
        self.assertIn("hello", converter.prefix_map)
        self.assertIn("CHEBI", converter.prefix_map)

    @SLOW
    def test_from_github(self) -> None:
        """Test getting a JSON-LD map from GitHub."""
        with self.assertRaises(ValueError):
            # missing end .jsonld file
            Converter.from_jsonld_github("biopragmatics", "bioregistry")

        semweb_converter = Converter.from_jsonld_github(
            "biopragmatics", "bioregistry", "exports", "contexts", "semweb.context.jsonld"
        )
        self.assertIn("rdf", semweb_converter.prefix_map)

    @SLOW
    def test_obo(self) -> None:
        """Test the OBO converter."""
        obo_converter = get_obo_converter()
        self.assertIn("CHEBI", obo_converter.prefix_map)
        self.assertNotIn("chebi", obo_converter.prefix_map)

    @SLOW
    def test_monarch(self) -> None:
        """Test the Monarch converter."""
        monarch_converter = get_monarch_converter()
        self.assertIn("CHEBI", monarch_converter.prefix_map)
        self.assertNotIn("chebi", monarch_converter.prefix_map)

    @SLOW
    def test_go_registry(self) -> None:
        """Test the GO registry converter."""
        go_converter = get_go_converter()
        self.assertIn("CHEBI", go_converter.prefix_map)
        self.assertNotIn("chebi", go_converter.prefix_map)

    def assert_bioregistry_converter(self, converter: Converter) -> None:
        """Assert the bioregistry converter has the right stuff in it."""
        records = {records.prefix: records for records in converter.records}
        self.assertIn("chebi", records)
        record = records["chebi"]
        self.assertIsInstance(record, Record)
        self.assertEqual("chebi", record.prefix)
        self.assertIn("CHEBI", record.prefix_synonyms)
        self.assertIn("ChEBI", record.prefix_synonyms)

        self.assertIn("chebi", converter.prefix_map)
        self.assertIn("chebi", converter.bimap)
        # Synonyms that are non-conflicting also get added
        self.assertIn("CHEBI", converter.prefix_map)
        self.assertNotIn("CHEBI", converter.bimap)
        chebi_uri = converter.prefix_map["chebi"]
        self.assertIn(chebi_uri, converter.reverse_prefix_map)
        self.assertEqual("chebi", converter.reverse_prefix_map[chebi_uri])

    def test_load_path(self) -> None:
        """Test loading from paths."""
        with tempfile.TemporaryDirectory() as directory:
            path = Path(directory).joinpath("pm.json")
            with self.assertRaises(FileNotFoundError):
                Converter.from_prefix_map(path)
            with self.assertRaises(FileNotFoundError):
                Converter.from_prefix_map(str(path))

            path.write_text(json.dumps(self.converter.prefix_map))

            c1 = Converter.from_prefix_map(path)
            self.assertEqual(self.converter.prefix_map, c1.prefix_map)

            c2 = Converter.from_prefix_map(str(path))
            self.assertEqual(self.converter.prefix_map, c2.prefix_map)

    def test_reverse_constructor(self) -> None:
        """Test constructing from a reverse prefix map."""
        converter = Converter.from_reverse_prefix_map(
            {
                "http://purl.obolibrary.org/obo/CHEBI_": "CHEBI",
                "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=": "CHEBI",
                "http://purl.obolibrary.org/obo/MONDO_": "MONDO",
            }
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488", converter.expand("CHEBI:138488")
        )

        self.assertEqual(
            "CHEBI:138488", converter.compress("http://purl.obolibrary.org/obo/CHEBI_138488")
        )
        self.assertEqual(
            "CHEBI:138488",
            converter.compress("https://www.ebi.ac.uk/chebi/searchId.do?chebiId=138488"),
        )

    def test_standardize_curie(self) -> None:
        """Test standardize CURIE."""
        converter = Converter.from_extended_prefix_map(
            [
                Record(
                    prefix="CHEBI",
                    prefix_synonyms=["chebi"],
                    uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                    uri_prefix_synonyms=[
                        "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:",
                    ],
                ),
            ]
        )
        self.assertEqual("CHEBI:138488", converter.standardize_curie("chebi:138488"))
        self.assertEqual("CHEBI:138488", converter.standardize_curie("CHEBI:138488"))
        self.assertIsNone(converter.standardize_curie("NOPE:NOPE"))
        self.assertEqual("NOPE:NOPE", converter.standardize_curie("NOPE:NOPE", passthrough=True))
        with self.assertRaises(CURIEStandardizationError):
            converter.standardize_curie("NOPE:NOPE", strict=True)

        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.standardize_uri(
                "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:138488"
            ),
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.standardize_uri("http://purl.obolibrary.org/obo/CHEBI_138488"),
        )
        self.assertIsNone(converter.standardize_uri("NOPE"))
        self.assertEqual("NOPE", converter.standardize_uri("NOPE", passthrough=True))
        with self.assertRaises(URIStandardizationError):
            converter.standardize_uri("NOPE:NOPE", strict=True)

    def test_combine(self) -> None:
        """Test chaining converters."""
        with self.assertRaises(ValueError):
            chain([])

        c1 = Converter.from_priority_prefix_map(
            {
                "CHEBI": ["http://purl.obolibrary.org/obo/CHEBI_", "https://bioregistry.io/chebi:"],
                "MONDO": ["http://purl.obolibrary.org/obo/MONDO_"],
            }
        )
        c2 = Converter.from_priority_prefix_map(
            {
                "CHEBI": [
                    "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=",
                    "http://identifiers.org/chebi/",
                    "http://purl.obolibrary.org/obo/CHEBI_",
                ],
                "GO": ["http://purl.obolibrary.org/obo/GO_"],
                "OBO": ["http://purl.obolibrary.org/obo/"],
            }
        )
        converter = chain([c1, c2], case_sensitive=True)

        self.assertEqual("CHEBI", converter.get_record("CHEBI").prefix)
        for url in [
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            "https://bioregistry.io/chebi:138488",
            "http://identifiers.org/chebi/138488",
            "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=138488",
        ]:
            self.assertEqual("CHEBI:138488", converter.compress(url))

        self.assertEqual("GO", converter.get_record("GO").prefix)
        self.assertEqual(
            "GO:0000001",
            converter.compress("http://purl.obolibrary.org/obo/GO_0000001"),
        )

        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_", converter.get_record("CHEBI").uri_prefix
        )
        self.assertIn("CHEBI", converter.prefix_map)
        self.assertEqual("http://purl.obolibrary.org/obo/CHEBI_", converter.prefix_map["CHEBI"])
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand("CHEBI:138488"),
        )
        self.assertNotIn("nope", converter.get_prefixes())

    def test_combine_with_synonyms(self) -> None:
        """Test combination with synonyms."""
        r1 = Record(prefix="GO", uri_prefix=GO_URI_PREFIX)
        r2 = Record(prefix="go", prefix_synonyms=["GO"], uri_prefix="https://identifiers.org/go:")

        c1 = Converter([])
        c1.add_record(r1)
        self.assertEqual(c1.records, Converter([r1]).records)

        c1.add_record(r2, merge=True)
        self.assertEqual(1, len(c1.records))
        r = c1.records[0]
        self.assertEqual("GO", r.prefix)
        self.assertEqual({"go"}, set(r.prefix_synonyms))
        self.assertEqual("http://purl.obolibrary.org/obo/GO_", r.uri_prefix)
        self.assertEqual({"https://identifiers.org/go:"}, set(r.uri_prefix_synonyms))

        c3 = chain([Converter([r1]), Converter([r2])])
        self.assertEqual(1, len(c3.records))
        self.assertIn("GO", c3.prefix_map)
        self.assertIn("go", c3.prefix_map, msg=f"PM: {c3.prefix_map}")
        self.assertNotIn("go", c3.bimap)
        self.assertIn("GO", c3.bimap)

    def test_combine_ci(self) -> None:
        """Test combining case-insensitive."""
        c1 = Converter.from_priority_prefix_map(
            {
                "CHEBI": [
                    "http://purl.obolibrary.org/obo/CHEBI_",
                    "https://bioregistry.io/chebi:",
                ],
            }
        )
        c2 = Converter.from_reverse_prefix_map(
            {
                "http://identifiers.org/chebi/": "chebi",
                "http://identifiers.org/chebi:": "chebi",
            }
        )
        converter = chain([c1, c2], case_sensitive=False)
        self.assertEqual({"CHEBI"}, converter.get_prefixes())
        self.assertEqual({"CHEBI"}, converter.get_prefixes(include_synonyms=False))
        self.assertEqual({"CHEBI", "chebi"}, converter.get_prefixes(include_synonyms=True))
        for url in [
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            "http://identifiers.org/chebi/138488",
            "http://identifiers.org/chebi:138488",
            "https://bioregistry.io/chebi:138488",
        ]:
            self.assertEqual("CHEBI:138488", converter.compress(url))
        # use the first prefix map for expansions
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand("CHEBI:138488"),
        )

    def test_combine_with_patterns(self) -> None:
        """Test chaining with patterns."""
        c1 = Converter([Record(prefix="a", uri_prefix="https://example.org/a/", pattern="^\\d{7}")])
        c2 = Converter([Record(prefix="a", uri_prefix="https://example.org/a/", pattern="^\\d+")])
        converter = chain([c1, c2])
        self.assertEqual(
            [Record(prefix="a", uri_prefix="https://example.org/a/", pattern="^\\d{7}")],
            converter.records,
        )

    def test_combine_with_patterns_via_synonym(self) -> None:
        """Test chaining with patterns."""
        c1 = Converter([Record(prefix="a", uri_prefix="https://example.org/a/", pattern="^\\d{7}")])
        c2 = Converter(
            [
                Record(
                    prefix="b",
                    prefix_synonyms=["a"],
                    uri_prefix="https://example.org/b/",
                    pattern="^\\d+",
                )
            ]
        )
        converter = chain([c1, c2])
        self.assertEqual(
            [
                Record(
                    prefix="a",
                    prefix_synonyms=["b"],
                    uri_prefix="https://example.org/a/",
                    uri_prefix_synonyms=["https://example.org/b/"],
                    pattern="^\\d{7}",
                )
            ],
            converter.records,
        )

    def test_df_bulk(self) -> None:
        """Test bulk processing in pandas dataframes."""
        rows = [
            ("CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
        ]
        df = pd.DataFrame(rows, columns=["curie", "uri"])
        self.converter.pd_expand(df, "curie")
        self.assertTrue((df.curie == df.uri).all())

        df = pd.DataFrame(rows, columns=["curie", "uri"])
        self.converter.pd_compress(df, "uri")
        self.assertTrue((df.curie == df.uri).all())

    def test_df_standardize(self) -> None:
        """Test standardizing dataframes."""
        converter = Converter([])
        converter.add_prefix(
            "chebi",
            "http://purl.obolibrary.org/obo/CHEBI_",
            prefix_synonyms=["CHEBI"],
            uri_prefix_synonyms=["https://bioregistry.io/chebi:"],
        )
        self.assertEqual("chebi", converter.standardize_prefix("chebi"))
        self.assertEqual("chebi", converter.standardize_prefix("CHEBI"))
        self.assertIsNone(converter.standardize_prefix("nope"))
        self.assertEqual("nope", converter.standardize_prefix("nope", passthrough=True))
        with self.assertRaises(PrefixStandardizationError):
            converter.standardize_prefix("nope", strict=True)

        rows = [
            ("chebi", "CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
            ("CHEBI", "CHEBI:2", "https://bioregistry.io/chebi:2"),
        ]
        df = pd.DataFrame(rows, columns=["prefix", "curie", "uri"])
        converter.pd_standardize_prefix(df, column="prefix")
        self.assertEqual(["chebi", "chebi"], list(df["prefix"]), msg=f"\n\n{df}")

        converter.pd_standardize_curie(df, column="curie")
        self.assertEqual(["chebi:1", "chebi:2"], list(df["curie"]))

        converter.pd_standardize_uri(df, column="uri")
        self.assertEqual(
            ["http://purl.obolibrary.org/obo/CHEBI_1", "http://purl.obolibrary.org/obo/CHEBI_2"],
            list(df["uri"]),
        )

    def test_file_bulk(self) -> None:
        """Test bulk processing of files."""
        with TemporaryDirectory() as directory:
            for rows, header in [
                (
                    [
                        ("curie", "uri"),
                        ("CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
                    ],
                    True,
                ),
                (
                    [
                        ("CHEBI:1", "http://purl.obolibrary.org/obo/CHEBI_1"),
                    ],
                    False,
                ),
            ]:
                path = Path(directory).joinpath("test.tsv")
                with path.open("w") as file:
                    for row in rows:
                        print(*row, sep="\t", file=file)

                idx = 1 if header else 0

                self.converter.file_expand(path, 0, header=header)
                lines = [line.strip().split("\t") for line in path.read_text().splitlines()]
                self.assertEqual("http://purl.obolibrary.org/obo/CHEBI_1", lines[idx][0])

                self.converter.file_compress(path, 0, header=header)
                lines = [line.strip().split("\t") for line in path.read_text().splitlines()]
                self.assertEqual("CHEBI:1", lines[idx][0])

    def test_incremental(self) -> None:
        """Test building a converter from an incremental interface."""
        converter = Converter([])
        for prefix, uri_prefix in self.simple_obo_prefix_map.items():
            converter.add_prefix(prefix, uri_prefix)
        converter.add_prefix(
            "hgnc",
            "https://bioregistry.io/hgnc:",
            prefix_synonyms=["HGNC"],
            uri_prefix_synonyms=["https://identifiers.org/hgnc:"],
        )
        self._assert_convert(converter)
        self.assertEqual(
            "hgnc:1234",
            converter.compress("https://bioregistry.io/hgnc:1234"),
        )
        self.assertEqual(
            "hgnc:1234",
            converter.compress("https://identifiers.org/hgnc:1234"),
        )
        self.assertEqual("https://bioregistry.io/hgnc:1234", converter.expand("HGNC:1234"))

        with self.assertRaises(ValueError):
            converter.add_prefix("GO", "...")
        with self.assertRaises(ValueError):
            converter.add_prefix("...", "http://purl.obolibrary.org/obo/GO_")
        with self.assertRaises(ValueError):
            converter.add_prefix(
                "...", "...", uri_prefix_synonyms=["http://purl.obolibrary.org/obo/GO_"]
            )
        with self.assertRaises(ValueError):
            converter.add_prefix("...", "...", prefix_synonyms=["GO"])

    def test_rdflib(self) -> None:
        """Test parsing a converter from an RDFLib object."""
        graph = rdflib.Graph()
        for prefix, uri_prefix in self.simple_obo_prefix_map.items():
            graph.bind(prefix, uri_prefix)
        converter = Converter.from_rdflib(graph)
        self._assert_convert(converter)

        converter_2 = Converter.from_rdflib(graph.namespace_manager)
        self._assert_convert(converter_2)

    def test_parse_curie(self) -> None:
        """Tests for parse CURIE."""
        converter = Converter(
            records=[
                Record(
                    prefix="GO",
                    uri_prefix="http://purl.obolibrary.org/obo/GO_",
                    prefix_synonyms=["go"],
                )
            ]
        )
        self.assertEqual(
            ReferenceTuple("GO", "1234567"), converter.parse_curie("GO:1234567", strict=True)
        )
        self.assertEqual(
            ReferenceTuple("GO", "1234567"), converter.parse_curie("GO:1234567", strict=False)
        )
        self.assertEqual(
            ReferenceTuple("GO", "1234567"), converter.parse_curie("go:1234567", strict=True)
        )
        self.assertEqual(
            ReferenceTuple("GO", "1234567"), converter.parse_curie("go:1234567", strict=False)
        )

        self.assertIsNone(converter.parse_curie("NOPE:NOPE", strict=False))
        with self.assertRaises(PrefixStandardizationError):
            converter.parse_curie("NOPE:NOPE", strict=True)

    def test_parse(self) -> None:
        """Test parsing URI or CURIE."""
        converter = Converter(
            records=[
                Record(
                    prefix="GO",
                    uri_prefix="http://purl.obolibrary.org/obo/GO_",
                    prefix_synonyms=["go"],
                    uri_prefix_synonyms=["https://identifiers.org/GO:"],
                )
            ]
        )
        curie = "GO:1234567"
        curie2 = "go:1234567"
        uri = "http://purl.obolibrary.org/obo/GO_1234567"
        uri2 = "https://identifiers.org/GO:1234567"

        for s in [curie, curie2, uri, uri2]:
            self.assertEqual(ReferenceTuple("GO", "1234567"), converter.parse(s, strict=True))
            self.assertEqual(ReferenceTuple("GO", "1234567"), converter.parse(s, strict=False))

        # test invalid CURIE parsing
        self.assertIsNone(converter.parse("NOPE:NOPE", strict=False))
        with self.assertRaises(ValueError):
            converter.parse_uri("NOPE:NOPE", strict=True)

        # test invalid URI parsing
        self.assertIsNone(converter.parse("https://example.org/nope", strict=False))
        with self.assertRaises(ValueError):
            converter.parse_uri("https://example.org/nope", strict=True)

        # test whatever's left
        self.assertIsNone(converter.parse("1234567", strict=False))
        with self.assertRaises(ValueError):
            converter.parse_uri("1234567", strict=True)

    def test_compress_or_standardize(self) -> None:
        """Test standardizing URI or CURIE."""
        converter = Converter(
            records=[
                Record(
                    prefix="GO",
                    uri_prefix="http://purl.obolibrary.org/obo/GO_",
                    prefix_synonyms=["go"],
                    uri_prefix_synonyms=["https://identifiers.org/GO:"],
                )
            ]
        )
        curie = "GO:1234567"
        curie2 = "go:1234567"
        uri = "http://purl.obolibrary.org/obo/GO_1234567"
        uri2 = "https://identifiers.org/GO:1234567"

        for s in [curie, curie2, uri, uri2]:
            self.assertEqual(
                "GO:1234567", converter.compress_or_standardize(s, strict=True, passthrough=True)
            )
            self.assertEqual(
                "GO:1234567", converter.compress_or_standardize(s, strict=False, passthrough=True)
            )
            self.assertEqual(
                "GO:1234567", converter.compress_or_standardize(s, strict=True, passthrough=False)
            )
            self.assertEqual(
                "GO:1234567", converter.compress_or_standardize(s, strict=False, passthrough=False)
            )

        self.assertIsNone(
            converter.compress_or_standardize("NOPE:NOPE", strict=False, passthrough=False)
        )
        self.assertEqual(
            "NOPE:NOPE",
            converter.compress_or_standardize("NOPE:NOPE", strict=False, passthrough=True),
        )
        with self.assertRaises(ValueError):
            converter.compress_or_standardize("NOPE:NOPE", strict=True, passthrough=True)
        with self.assertRaises(ValueError):
            converter.compress_or_standardize("NOPE:NOPE", strict=True, passthrough=False)

    def test_parse_uri(self) -> None:
        """Tests for parsing URIs."""
        converter = Converter(
            records=[
                Record(
                    prefix="GO",
                    uri_prefix="http://purl.obolibrary.org/obo/GO_",
                    prefix_synonyms=["go"],
                    uri_prefix_synonyms=["https://identifiers.org/GO:"],
                )
            ]
        )
        uri = "http://purl.obolibrary.org/obo/GO_1234567"
        uri2 = "https://identifiers.org/GO:1234567"

        self.assertEqual(ReferenceTuple("GO", "1234567"), converter.parse_uri(uri, strict=True))
        self.assertEqual(ReferenceTuple("GO", "1234567"), converter.parse_uri(uri, strict=False))

        self.assertEqual(ReferenceTuple("GO", "1234567"), converter.parse_uri(uri2, strict=True))
        self.assertEqual(ReferenceTuple("GO", "1234567"), converter.parse_uri(uri2, strict=False))

        self.assertEqual(
            (None, None), converter.parse_uri("123345", strict=False, return_none=False)
        )
        self.assertIsNone(converter.parse_uri("123345", strict=False, return_none=True))
        with self.assertRaises(ValueError):
            converter.parse_uri("123345", strict=True)

    def test_expand(self) -> None:
        """Tests for expand."""
        converter = Converter(
            records=[
                Record(
                    prefix="GO",
                    uri_prefix="http://purl.obolibrary.org/obo/GO_",
                    prefix_synonyms=["go"],
                )
            ]
        )
        uri = "http://purl.obolibrary.org/obo/GO_1234567"
        self.assertEqual(uri, converter.expand("GO:1234567", strict=True, passthrough=True))
        self.assertEqual(uri, converter.expand("GO:1234567", strict=False, passthrough=True))
        self.assertEqual(uri, converter.expand("GO:1234567", strict=True, passthrough=False))
        self.assertEqual(uri, converter.expand("GO:1234567", strict=False, passthrough=False))

        self.assertEqual(uri, converter.expand("go:1234567", strict=True, passthrough=True))
        self.assertEqual(uri, converter.expand("go:1234567", strict=False, passthrough=True))
        self.assertEqual(uri, converter.expand("go:1234567", strict=True, passthrough=False))
        self.assertEqual(uri, converter.expand("go:1234567", strict=False, passthrough=False))

        self.assertEqual("NOPE:NOPE", converter.expand("NOPE:NOPE", strict=False, passthrough=True))
        self.assertIsNone(converter.expand("NOPE:NOPE", strict=False, passthrough=False))

        with self.assertRaises(ExpansionError):
            converter.expand("NOPE:NOPE", strict=True, passthrough=True)
        with self.assertRaises(ExpansionError):
            converter.expand("NOPE:NOPE", strict=True, passthrough=False)

    def test_expand_pair_all(self) -> None:
        """Tests for expand."""
        converter = Converter(
            records=[
                Record(
                    prefix="GO",
                    uri_prefix="http://purl.obolibrary.org/obo/GO_",
                    prefix_synonyms=["go"],
                    uri_prefix_synonyms=["https://identifiers.org/GO:"],
                )
            ]
        )
        uris = ["http://purl.obolibrary.org/obo/GO_1234567", "https://identifiers.org/GO:1234567"]
        self.assertEqual(uris, converter.expand_pair_all("GO", "1234567", strict=True))
        self.assertEqual(uris, converter.expand_pair_all("GO", "1234567", strict=False))

        # with synonym as input
        self.assertEqual(uris, converter.expand_pair_all("go", "1234567", strict=True))
        self.assertEqual(uris, converter.expand_pair_all("go", "1234567", strict=False))

        self.assertIsNone(converter.expand_pair_all("NOPE", "NOPE", strict=False))
        with self.assertRaises(ExpansionError):
            converter.expand_pair_all("NOPE", "NOPE", strict=True)

    def test_expand_reference(self) -> None:
        """Tests for expand."""
        converter = Converter(
            records=[
                Record(
                    prefix="GO",
                    uri_prefix="http://purl.obolibrary.org/obo/GO_",
                    prefix_synonyms=["go"],
                )
            ]
        )
        uri = "http://purl.obolibrary.org/obo/GO_1234567"
        ref = ReferenceTuple("GO", "1234567")
        ref2 = ReferenceTuple("go", "1234567")
        nope = ReferenceTuple("NOPE", "NOPE")
        self.assertEqual(uri, converter.expand_reference(ref, strict=True, passthrough=True))
        self.assertEqual(uri, converter.expand_reference(ref, strict=False, passthrough=True))
        self.assertEqual(uri, converter.expand_reference(ref, strict=True, passthrough=False))
        self.assertEqual(uri, converter.expand_reference(ref, strict=False, passthrough=False))

        self.assertEqual(uri, converter.expand_reference(ref2, strict=True, passthrough=True))
        self.assertEqual(uri, converter.expand_reference(ref2, strict=False, passthrough=True))
        self.assertEqual(uri, converter.expand_reference(ref2, strict=True, passthrough=False))
        self.assertEqual(uri, converter.expand_reference(ref2, strict=False, passthrough=False))

        self.assertEqual(
            "NOPE:NOPE", converter.expand_reference(nope, strict=False, passthrough=True)
        )
        self.assertIsNone(converter.expand_reference(nope, strict=False, passthrough=False))

        with self.assertRaises(ExpansionError):
            converter.expand_reference(nope, strict=True, passthrough=True)
        with self.assertRaises(ExpansionError):
            converter.expand_reference(nope, strict=True, passthrough=False)

    def test_expand_all(self) -> None:
        """Test expand all."""
        priority_prefix_map = {
            "CHEBI": [
                "http://purl.obolibrary.org/obo/CHEBI_",
                "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:",
            ],
        }
        converter = Converter.from_priority_prefix_map(priority_prefix_map)
        self.assertEqual(
            [
                "http://purl.obolibrary.org/obo/CHEBI_138488",
                "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=CHEBI:138488",
            ],
            converter.expand_all("CHEBI:138488"),
        )
        self.assertIsNone(converter.expand_all("NOPE:NOPE"))
        with self.assertRaises(PrefixStandardizationError):
            converter.expand_all("NOPE:NOPE", strict=True)

    def test_expand_ambiguous(self) -> None:
        """Test expansion of URI or CURIEs."""
        converter = Converter.from_extended_prefix_map(
            [
                Record(
                    prefix="CHEBI",
                    prefix_synonyms=["chebi"],
                    uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                    uri_prefix_synonyms=["https://identifiers.org/chebi:"],
                ),
            ]
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand_or_standardize("CHEBI:138488"),
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand_or_standardize("chebi:138488"),
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand_or_standardize("http://purl.obolibrary.org/obo/CHEBI_138488"),
        )
        self.assertEqual(
            "http://purl.obolibrary.org/obo/CHEBI_138488",
            converter.expand_or_standardize("https://identifiers.org/chebi:138488"),
        )

        self.assertIsNone(converter.expand_or_standardize("missing:0000000"))
        with self.assertRaises(ExpansionError):
            converter.expand_or_standardize("missing:0000000", strict=True)
        self.assertEqual(
            "missing:0000000", converter.expand_or_standardize("missing:0000000", passthrough=True)
        )

        self.assertIsNone(converter.expand_or_standardize("https://example.com/missing:0000000"))
        with self.assertRaises(ExpansionError):
            converter.expand_or_standardize("https://example.com/missing:0000000", strict=True)
        self.assertEqual(
            "https://example.com/missing:0000000",
            converter.expand_or_standardize(
                "https://example.com/missing:0000000", passthrough=True
            ),
        )

    def test_compress_ambiguous(self) -> None:
        """Test compression of URI or CURIEs."""
        converter = Converter.from_extended_prefix_map(
            [
                Record(
                    prefix="CHEBI",
                    prefix_synonyms=["chebi"],
                    uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                    uri_prefix_synonyms=["https://identifiers.org/chebi:"],
                ),
            ]
        )
        self.assertEqual("CHEBI:138488", converter.compress_or_standardize("CHEBI:138488"))
        self.assertEqual("CHEBI:138488", converter.compress_or_standardize("chebi:138488"))
        self.assertEqual(
            "CHEBI:138488",
            converter.compress_or_standardize("http://purl.obolibrary.org/obo/CHEBI_138488"),
        )
        self.assertEqual(
            "CHEBI:138488",
            converter.compress_or_standardize("https://identifiers.org/chebi:138488"),
        )

        self.assertIsNone(converter.expand_or_standardize("missing:0000000"))
        with self.assertRaises(ExpansionError):
            converter.expand_or_standardize("missing:0000000", strict=True)
        self.assertEqual(
            "missing:0000000", converter.expand_or_standardize("missing:0000000", passthrough=True)
        )

        self.assertIsNone(converter.expand_or_standardize("https://example.com/missing:0000000"))
        with self.assertRaises(ExpansionError):
            converter.expand_or_standardize("https://example.com/missing:0000000", strict=True)
        self.assertEqual(
            "https://example.com/missing:0000000",
            converter.expand_or_standardize(
                "https://example.com/missing:0000000", passthrough=True
            ),
        )

    def test_standardize_identifier(self) -> None:
        """Test standardizing identifiers."""

        class BananaStripperConverter(Converter):
            """A converter that removes bananas from LUIDs."""

            def standardize_identifier(self, prefix: str, identifier: str) -> str | None:
                """Standardize the identifier by removing a banana and checking it is numeric."""
                norm_identifier = identifier.removeprefix(f"{prefix}:")

                # now, do some validation
                if not norm_identifier.isnumeric():
                    return None

                return norm_identifier

        converter = BananaStripperConverter(
            records=[
                Record(
                    prefix="CHEBI",
                    prefix_synonyms=["chebi"],
                    uri_prefix="http://purl.obolibrary.org/obo/CHEBI_",
                    uri_prefix_synonyms=["https://identifiers.org/chebi:"],
                ),
            ]
        )
        self.assertEqual(ReferenceTuple("CHEBI", "1234"), converter.parse_curie("CHEBI:1234"))
        self.assertEqual(ReferenceTuple("CHEBI", "1234"), converter.parse_curie("chebi:1234"))
        self.assertEqual(ReferenceTuple("CHEBI", "1234"), converter.parse_curie("CHEBI:CHEBI:1234"))
        self.assertEqual(ReferenceTuple("CHEBI", "1234"), converter.parse_curie("chebi:CHEBI:1234"))
        self.assertIsNone(converter.parse_curie("NOPE:NOPE:1234", strict=False))

        self.assertIsNone(converter.parse_curie("CHEBI:nope", strict=False))
        # does not solve the problem of synonyms in the banana, this is specific
        # to the current implementation in this test
        self.assertIsNone(converter.parse_curie("chebi:chebi:1234", strict=False))
        with self.assertRaises(ValueError):
            converter.parse_curie("CHEBI:nope", strict=True)


class TestVersion(unittest.TestCase):
    """Trivially test a version."""

    def test_version_type(self) -> None:
        """Test the version is a string.

        This is only meant to be an example test.
        """
        version = get_version()
        self.assertIsInstance(version, str)


class TestUtils(unittest.TestCase):
    """Test utility functions."""

    def test_clean(self) -> None:
        """Test clean."""
        prefix_map = {
            "b": "https://example.com/a/",
            "a": "https://example.com/a/",
            "c": "https://example.com/c/",
        }
        records = upgrade_prefix_map(prefix_map)
        self.assertEqual(2, len(records))
        a_record, c_record = records

        self.assertEqual("a", a_record.prefix)
        self.assertEqual(["b"], a_record.prefix_synonyms)
        self.assertEqual("https://example.com/a/", a_record.uri_prefix)
        self.assertEqual([], a_record.uri_prefix_synonyms)

        self.assertEqual("c", c_record.prefix)
        self.assertEqual([], c_record.prefix_synonyms)
        self.assertEqual("https://example.com/c/", c_record.uri_prefix)
        self.assertEqual([], c_record.uri_prefix_synonyms)



"""Utilities for triples."""

from __future__ import annotations

import csv
import gzip
from collections.abc import Generator, Iterable, Sequence
from contextlib import contextmanager
from pathlib import Path
from typing import NamedTuple, TextIO

from pydantic import BaseModel, ConfigDict
from typing_extensions import Self

from .api import Reference

__all__ = [
    "StrTriple",
    "Triple",
    "read_triples",
    "write_triples",
]


class StrTriple(NamedTuple):
    """A triple of curies."""

    subject: str
    predicate: str
    object: str


class Triple(BaseModel):
    """A model for a triple of subject-predicate-object triple."""

    model_config = ConfigDict(frozen=True)

    subject: Reference
    predicate: Reference
    object: Reference

    def as_str_triple(self) -> StrTriple:
        """Get a three-tuple of strings representing this triple."""
        return StrTriple(self.subject.curie, self.predicate.curie, self.object.curie)

    def __lt__(self, other: Triple) -> bool:
        return self.as_str_triple() < other.as_str_triple()

    @classmethod
    def from_curies(
        cls,
        subject_curie: str,
        predicate_curie: str,
        object_curie: str,
        *,
        reference_cls: type[Reference] = Reference,
    ) -> Self:
        """Construct a triple from three CURIE strings."""
        return cls(
            subject=reference_cls.from_curie(subject_curie),
            predicate=reference_cls.from_curie(predicate_curie),
            object=reference_cls.from_curie(object_curie),
        )


#: the default header for a three-column file representing triples
HEADER = list(Triple.model_fields)


@contextmanager
def _get_file(path: str | Path, read: bool) -> Generator[TextIO, None, None]:
    path = Path(path).expanduser().resolve()
    if path.suffix == ".gz":
        yield gzip.open(path, mode="rt" if read else "wt")
    else:
        yield open(path, mode="r" if read else "w")


def write_triples(
    triples: Iterable[Triple], path: str | Path, *, header: Sequence[str] | None = None
) -> None:
    """Write triples to a file."""
    if header is None:
        header = HEADER
    with _get_file(path, read=False) as file:
        writer = csv.writer(file, delimiter="\t")
        writer.writerow(header)
        writer.writerows(
            (triple.subject.curie, triple.predicate.curie, triple.object.curie)
            for triple in triples
        )


def read_triples(path: str | Path, *, reference_cls: type[Reference] | None = None) -> list[Triple]:
    """Read triples."""
    if reference_cls is None:
        reference_cls = Reference
    with _get_file(path, read=True) as file:
        reader = csv.reader(file, delimiter="\t")
        _header = next(reader)
        return [
            Triple(
                subject=reference_cls.from_curie(subject_curie),
                predicate=reference_cls.from_curie(predicate_curie),
                object=reference_cls.from_curie(object_curie),
            )
            for subject_curie, predicate_curie, object_curie in reader
        ]



"""Identifier mappings service.

This contains an implementation of the service described in `SPARQL-enabled identifier
conversion with Identifiers.org <https://pubmed.ncbi.nlm.nih.gov/25638809/>`_. The idea
here is that you can write a SPARQL query like the following:

.. code-block:: sparql

    PREFIX biomodel: <http://identifiers.org/biomodels.db/>
    PREFIX bqbio: <http://biomodels.net/biology-qualifiers#>
    PREFIX sbmlrdf: <http://identifiers.org/biomodels.vocabulary#>
    PREFIX up: <http://purl.uniprot.org/core/>

    SELECT DISTINCT ?protein ?protein_domain
    WHERE {
        # The first part of this query extracts the proteins appearing in an RDF serialization
        # of the BioModels database (see https://www.ebi.ac.uk/biomodels/BIOMD0000000372) on
        # insulin/glucose feedback. Note that modelers call entities appearing in compartmental
        # models "species", and this does not refer to taxa.
        biomodel:BIOMD0000000372 sbmlrdf:species/bqbio:isVersionOf ?biomodels_protein .

        # The second part of this query maps BioModels protein IRIs to UniProt protein IRIs
        # using service XXX - that's what we're implementing here.
        SERVICE <XXX> {
            ?biomodels_protein owl:sameAs ?uniprot_protein.
        }

        # The third part of this query gets links between UniProt proteins and their
        # domains. Since the service maps between the BioModels query, this only gets
        # us relevant protein domains to the insulin/glucose model.
        SERVICE <http://beta.sparql.uniprot.org/sparql> {
            ?uniprot_protein a up:Protein;
                up:organism taxon:9606;
                rdfs:seeAlso ?protein_domain.
        }
    }

The SPARQL endpoint running at the web address XXX takes in the bound values for
`?biomodels_protein` one at a time and dynamically generates triples with `owl:sameAs`
as the predicate mapping and other equivalent IRIs (based on the definition of the
converter) as the objects. This allows for gluing together multiple services that use
different URIs for the same entities - in this example, there are two ways of referring
to UniProt Proteins:

1. The BioModels database example represents a SBML model on insulin-glucose feedback
   and uses legacy Identifiers.org URIs for proteins such as
   http://identifiers.org/uniprot/P01308.
2. The first-part UniProt database uses its own PURLs such as
   https://purl.uniprot.org/uniprot/P01308.

.. seealso::

    - Jerven Bolleman's implementation of this service in Java:
      https://github.com/JervenBolleman/sparql-identifiers
    - Vincent Emonet's `SPARQL endpoint for RDFLib generator
      <https://github.com/vemonet/rdflib-endpoint>`_

The following is an end-to-end example of using this function to create a small URI
mapping application.

.. code-block:: python

    # flask_example.py
    from flask import Flask
    from curies import Converter, get_bioregistry_converter
    from curies.mapping_service import get_flask_mapping_app

    # Create a converter
    converter: Converter = get_bioregistry_converter()

    # Create the Flask app from the converter
    app: Flask = get_flask_mapping_app(converter)

    if __name__ == "__main__":
        app.run()

In the command line, either run your Python file directly, or via with :mod:`gunicorn`:

.. code-block:: console

    $ pip install gunicorn
    $ gunicorn --bind 0.0.0.0:8764 flask_example:app

Test a request in the Python REPL.

.. code-block:: python

    import requests

    sparql = '''
        SELECT ?s ?o WHERE {
            VALUES ?s { <http://purl.obolibrary.org/obo/CHEBI_2> }
            ?s owl:sameAs ?o
        }
    '''
    res = requests.get("http://localhost:8764/sparql", params={"query": sparql})

Test a request using a service, e.g. with :meth:`rdflib.Graph.query`

.. code-block:: sparql

    SELECT ?s ?o WHERE {
        VALUES ?s { <http://purl.obolibrary.org/obo/CHEBI_2> }
        SERVICE <http://localhost:8764/sparql> {
            ?s owl:sameAs ?child_mapped .
        }
    }
"""

from .api import (
    MappingServiceGraph,
    get_fastapi_mapping_app,
    get_fastapi_router,
    get_flask_mapping_app,
    get_flask_mapping_blueprint,
)
from .rdflib_custom import MappingServiceSPARQLProcessor  # type:ignore

__all__ = [
    "MappingServiceGraph",
    "MappingServiceSPARQLProcessor",
    "get_fastapi_mapping_app",
    "get_fastapi_router",
    "get_flask_mapping_app",
    "get_flask_mapping_blueprint",
]



"""Tests for the identifier mapping service."""

import unittest
from typing import Union
from urllib.parse import quote

import httpx
import werkzeug.test
from fastapi.testclient import TestClient
from flask.testing import FlaskClient
from rdflib import OWL, SKOS
from rdflib.query import Result

from curies import Converter
from curies.mapping_service import (
    MappingServiceGraph,
    MappingServiceSPARQLProcessor,
    get_fastapi_mapping_app,
    get_flask_mapping_app,
)
from curies.mapping_service.api import _prepare_predicates
from curies.mapping_service.utils import (
    CONTENT_TYPE_SYNONYMS,
    CONTENT_TYPE_TO_HANDLER,
    handle_header,
    sparql_service_available,
)
from tests.constants import SLOW

VALID_CONTENT_TYPES = {
    *CONTENT_TYPE_TO_HANDLER,
    "",
    "*/*",
    *CONTENT_TYPE_SYNONYMS,
}

PREFIX_MAP = {
    "CHEBI": [
        "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=",
        "http://identifiers.org/chebi/",
        "http://purl.obolibrary.org/obo/CHEBI_",
    ],
    "GO": ["http://purl.obolibrary.org/obo/GO_"],
    "OBO": ["http://purl.obolibrary.org/obo/"],
}

SPARQL_SIMPLE = """\
SELECT DISTINCT ?s ?o WHERE {
    VALUES ?s {
        <http://purl.obolibrary.org/obo/CHEBI_1>
        <http://purl.obolibrary.org/obo/CHEBI_2>
    }
    ?s owl:sameAs ?o
}
""".rstrip()

SPARQL_SIMPLE_BACKWARDS = """\
SELECT DISTINCT ?s ?o WHERE {
    VALUES ?o {
        <http://purl.obolibrary.org/obo/CHEBI_1>
        <http://purl.obolibrary.org/obo/CHEBI_2>
    }
    ?s owl:sameAs ?o
}
""".rstrip()

#: This represents a SPARQL query that happens when a service generates it
SPARQL_FROM_SERVICE = """\
SELECT REDUCED * WHERE {
    ?s owl:sameAs ?o .
}
VALUES (?s) {
    (<http://purl.obolibrary.org/obo/CHEBI_1>)
    (<http://purl.obolibrary.org/obo/CHEBI_2>)
}
"""

EXPECTED = {
    (
        "http://purl.obolibrary.org/obo/CHEBI_1",
        "http://purl.obolibrary.org/obo/CHEBI_1",
    ),
    ("http://purl.obolibrary.org/obo/CHEBI_1", "http://identifiers.org/chebi/1"),
    (
        "http://purl.obolibrary.org/obo/CHEBI_1",
        "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=1",
    ),
    (
        "http://purl.obolibrary.org/obo/CHEBI_2",
        "http://purl.obolibrary.org/obo/CHEBI_2",
    ),
    ("http://purl.obolibrary.org/obo/CHEBI_2", "http://identifiers.org/chebi/2"),
    (
        "http://purl.obolibrary.org/obo/CHEBI_2",
        "https://www.ebi.ac.uk/chebi/searchId.do?chebiId=2",
    ),
}


def _stm(rows: Result) -> set[tuple[str, str]]:
    return {(str(row.s), str(row.o)) for row in rows}  # type: ignore


class TestMappingService(unittest.TestCase):
    """Test the identifier mapping service."""

    def setUp(self) -> None:
        """Set up the converter."""
        self.converter = Converter.from_priority_prefix_map(PREFIX_MAP)
        self.graph = MappingServiceGraph(converter=self.converter)
        self.processor = MappingServiceSPARQLProcessor(self.graph)

    def test_parse_header(self) -> None:
        """Test parsing a rather complex header."""
        example_header = (
            "application/sparql-results+xml;q=0.8,"
            "application/xml;q=0.8,"
            "application/x-binary-rdf-results-table,"
            "application/sparql-results+json;q=0.8,"
            "application/json;q=0.8,"
            "text/csv;q=0.8,"
            "text/tab-separated-values;q=0.8"
        )
        content_type = handle_header(example_header)
        self.assertEqual("application/sparql-results+xml", content_type)

    def test_prepare_predicates(self) -> None:
        """Test preparation of predicates."""
        self.assertEqual({OWL.sameAs}, _prepare_predicates())
        self.assertEqual({OWL.sameAs}, _prepare_predicates(OWL.sameAs))
        self.assertEqual(
            {OWL.sameAs, SKOS.exactMatch}, _prepare_predicates({OWL.sameAs, SKOS.exactMatch})
        )

    def test_errors(self) -> None:
        """Test errors."""
        for sparql in [
            # errors because of unbound subject
            "SELECT ?s ?o WHERE { ?s owl:sameAs ?o }",
            # errors because of bad predicate
            "SELECT ?o WHERE { <http://purl.obolibrary.org/obo/CHEBI_1> rdfs:seeAlso ?o }",
            "SELECT ?s WHERE { ?s rdfs:seeAlso <http://purl.obolibrary.org/obo/CHEBI_1> }",
            # errors because of unknown URI
            "SELECT ?o WHERE { <http://example.com/1> owl:sameAs ?o }",
            "SELECT ?s WHERE { ?s owl:sameAs <http://example.com/1> }",
            # errors because predicate is given
            "SELECT * WHERE { <http://purl.obolibrary.org/obo/CHEBI_1> "
            "owl:sameAs <http://purl.obolibrary.org/obo/CHEBI_1> }",
        ]:
            with self.subTest(sparql=sparql):
                self.assertEqual([], list(self.graph.query(sparql, processor=self.processor)))

    def test_sparql(self) -> None:
        """Test a sparql query on the graph."""
        rows = _stm(self.graph.query(SPARQL_SIMPLE, processor=self.processor))
        self.assertNotEqual(0, len(rows), msg="No results were returned")
        self.assertEqual(EXPECTED, rows)

    def test_sparql_backwards(self) -> None:
        """Test a sparql query on the graph."""
        rows = _stm(self.graph.query(SPARQL_SIMPLE_BACKWARDS, processor=self.processor))
        self.assertNotEqual(0, len(rows), msg="No results were returned")
        expected = {(o, s) for s, o in EXPECTED}
        self.assertEqual(expected, rows)

    def test_service_sparql(self) -> None:
        """Test the SPARQL that gets sent when using this as a service."""
        rows = _stm(self.graph.query(SPARQL_FROM_SERVICE, processor=self.processor))
        self.assertNotEqual(0, len(rows), msg="No results were returned")
        self.assertEqual(EXPECTED, rows)

    def test_missing(self) -> None:
        """Test a sparql query on the graph where the URIs can't be parsed."""
        sparql = """\
            SELECT ?s ?o WHERE {
                VALUES ?s { <http://example.org/1> <http://example.org/1> }
                ?s owl:sameAs ?o
            }
        """
        self.assertEqual([], list(self.graph.query(sparql, processor=self.processor)))

    def test_safe_expand(self) -> None:
        """Test that expansion to invalid prefixes doesn't happen."""
        ppm = {
            "CHEBI": [
                "http://purl.obolibrary.org/obo/CHEBI_",
                "http://identifiers.org/chebi/",
                "http://identifiers.org/chebi/nope nope:",
            ],
        }
        converter = Converter.from_priority_prefix_map(ppm)
        graph = MappingServiceGraph(converter=converter)
        self.assertEqual(
            {"http://purl.obolibrary.org/obo/CHEBI_1", "http://identifiers.org/chebi/1"},
            set(map(str, graph._expand_pair_all("http://purl.obolibrary.org/obo/CHEBI_1"))),
        )


class ConverterMixin(unittest.TestCase):
    """A mixin that has a converter."""

    def setUp(self) -> None:
        """Set up the test case with a converter."""
        super().setUp()
        self.converter = Converter.from_priority_prefix_map(PREFIX_MAP)

    def assert_mimetype(
        self, res: Union[httpx.Response, werkzeug.test.TestResponse], content_type: str
    ) -> None:
        """Assert the correct MIMETYPE."""
        content_type = handle_header(content_type)
        mimetype = getattr(res, "mimetype", None)
        if hasattr(res, "mimetype"):  # this is from Flask
            self.assertEqual(content_type, mimetype)
        else:  # this is from FastAPI
            actual_content_type = res.headers.get("content-type")
            self.assertIsNotNone(actual_content_type)
            self.assertEqual(content_type, actual_content_type.split(";")[0].strip())

    def assert_parsed(
        self, res: Union[httpx.Response, werkzeug.test.TestResponse], content_type: str
    ) -> None:
        """Test the result has the expected output."""
        content_type = handle_header(content_type)
        parse_func = CONTENT_TYPE_TO_HANDLER[content_type]
        records = parse_func(res.text)
        pairs = {(record["s"], record["o"]) for record in records}
        self.assertEqual(EXPECTED, pairs)

    def assert_get_sparql_results(
        self, client: Union[TestClient, FlaskClient], sparql: str
    ) -> None:
        """Test a sparql query returns expected values."""
        for content_type in sorted(VALID_CONTENT_TYPES):
            with self.subTest(content_type=content_type):
                res = client.get(f"/sparql?query={quote(sparql)}", headers={"accept": content_type})
                self.assertEqual(200, res.status_code, msg=f"Response: {res}\n\n{res.text}")
                self.assert_mimetype(res, content_type)
                self.assert_parsed(res, content_type)

    def assert_post_sparql_results(
        self, client: Union[TestClient, FlaskClient], sparql: str
    ) -> None:
        """Test a sparql query returns expected values."""
        for content_type in sorted(VALID_CONTENT_TYPES):
            with self.subTest(content_type=content_type):
                res = client.post(
                    # note that we're using "data" and not JSON since this service
                    # is posting "form data" and not a JSON payload
                    "/sparql",
                    data={"query": sparql},
                    headers={"accept": content_type},
                )
                self.assertEqual(
                    200,
                    res.status_code,
                    msg=f"Response: {res}",
                )
                self.assert_mimetype(res, content_type)
                self.assert_parsed(res, content_type)


class TestFlaskMappingWeb(ConverterMixin):
    """Test the Flask-based mapping service."""

    def setUp(self) -> None:
        """Set up the test case with a converter and app."""
        super().setUp()
        self.app = get_flask_mapping_app(self.converter)

    def test_get_missing_query(self) -> None:
        """Test error on missing query parameter."""
        with self.app.test_client() as client:
            for content_type in sorted(VALID_CONTENT_TYPES):
                with self.subTest(content_type=content_type):
                    res = client.get("/sparql", headers={"accept": content_type})
                    self.assertEqual(400, res.status_code, msg=f"Response: {res}")

    def test_post_missing_query(self) -> None:
        """Test error on missing query parameter."""
        with self.app.test_client() as client:
            for content_type in sorted(VALID_CONTENT_TYPES):
                with self.subTest(content_type=content_type):
                    res = client.post("/sparql", headers={"accept": content_type})
                    self.assertEqual(400, res.status_code, msg=f"Response: {res}")

    def test_get_query(self) -> None:
        """Test querying the app with GET."""
        with self.app.test_client() as client:
            self.assert_get_sparql_results(client, SPARQL_SIMPLE)

    def test_post_query(self) -> None:
        """Test querying the app with POST."""
        with self.app.test_client() as client:
            self.assert_post_sparql_results(client, SPARQL_SIMPLE)

    def test_get_service_query(self) -> None:
        """Test sparql generated by a service (that has values outside of where clause) with GET."""
        with self.app.test_client() as client:
            self.assert_get_sparql_results(client, SPARQL_FROM_SERVICE)

    def test_post_service_query(self) -> None:
        """Test sparql generated by a service (that has values outside of where clause) with POST."""
        with self.app.test_client() as client:
            self.assert_post_sparql_results(client, SPARQL_FROM_SERVICE)


class TestFastAPIMappingApp(ConverterMixin):
    """Test the FastAPI-based mapping service."""

    def setUp(self) -> None:
        """Set up the test case with a converter, blueprint, and app."""
        super().setUp()
        self.app = get_fastapi_mapping_app(self.converter)
        self.client = TestClient(self.app)

    def test_get_missing_query(self) -> None:
        """Test error on missing query parameter."""
        for content_type in sorted(VALID_CONTENT_TYPES):
            with self.subTest(content_type=content_type):
                res = self.client.get("/sparql", headers={"accept": content_type})
                self.assertEqual(422, res.status_code, msg=f"Response: {res}")

    def test_post_missing_query(self) -> None:
        """Test error on missing query parameter."""
        for content_type in sorted(VALID_CONTENT_TYPES):
            with self.subTest(content_type=content_type):
                res = self.client.post("/sparql", headers={"accept": content_type})
                self.assertEqual(422, res.status_code, msg=f"Response: {res}")

    @unittest.skip(reason="Weird failures on CI")
    def test_get_query(self) -> None:
        """Test querying the app with GET."""
        self.assert_get_sparql_results(self.client, SPARQL_SIMPLE)

    def test_post_query(self) -> None:
        """Test querying the app with POST."""
        self.assert_post_sparql_results(self.client, SPARQL_SIMPLE)

    @unittest.skip(reason="Weird failures on CI")
    def test_get_service_query(self) -> None:
        """Test sparql generated by a service (that has values outside of where clause) with GET."""
        self.assert_get_sparql_results(self.client, SPARQL_FROM_SERVICE)

    def test_post_service_query(self) -> None:
        """Test sparql generated by a service (that has values outside of where clause) with POST."""
        self.assert_post_sparql_results(self.client, SPARQL_FROM_SERVICE)


class TestUtils(unittest.TestCase):
    """Test utilities."""

    @SLOW
    def test_availability(self) -> None:
        """Test sparql service availability check."""
        self.assertTrue(
            sparql_service_available("https://query.wikidata.org/bigdata/namespace/wdq/sparql")
        )
        self.assertFalse(sparql_service_available("https://example.org"))



"""External sources of contexts."""

from __future__ import annotations

from typing import Any

from .api import Converter

__all__ = [
    "get_bioregistry_converter",
    "get_go_converter",
    "get_monarch_converter",
    "get_obo_converter",
    "get_prefixcommons_converter",
]

BIOREGISTRY_CONTEXTS = (
    "https://raw.githubusercontent.com/biopragmatics/bioregistry/main/exports/contexts"
)


def get_obo_converter() -> Converter:
    """Get the latest OBO Foundry context.

    :returns:
        A converter object representing the OBO Foundry's JSON-LD context,
        which contains a simple mapping from OBO Foundry preferred prefixes
        for ontologies that contain case stylization (e.g., ``GO``, not ``go``; ``VariO``, not ``vario``).

        It does not include synonyms nor any non-ontology prefixes - e.g., it does not include
        semantic web prefixes like ``rdfs``, it does not include other useful biomedical prefixes
        like ``hgnc``.

    If you want a more comprehensive prefix map, consider using the Bioregistry
    via :func:`get_bioregistry_converter` or by chaining the OBO converter in front of the
    Bioregistry depending on your personal/project preferences using :func:`curies.chain`.

    Provenance:

    - This JSON-LD context is generated programmatically
      by https://github.com/OBOFoundry/OBOFoundry.github.io/blob/master/util/processor.py.
    - The file is accessed via from http://purl.obolibrary.org/meta/obo_context.jsonld,
      which is configured through the OBO Foundry's PURL server with
      https://github.com/OBOFoundry/purl.obolibrary.org/blob/master/www/.htaccess
      and ultimately points to
      https://raw.githubusercontent.com/OBOFoundry/OBOFoundry.github.io/master/registry/obo_context.jsonl
    """
    # See configuration on
    # to see where this PURL points
    url = "http://purl.obolibrary.org/meta/obo_context.jsonld"
    return Converter.from_jsonld(url)


def get_prefixcommons_converter(name: str = "monarch_context") -> Converter:
    """Get a Prefix Commons-maintained context.

    :param name: The name of the JSON-LD file (e.g., defaults to ``monarch_context``).
        See the full list at https://github.com/prefixcommons/prefixcommons-py/tree/master/prefixcommons/registry.
    :returns:
        A converter
    """
    url = (
        "https://raw.githubusercontent.com/prefixcommons/prefixcommons-py/master/"
        f"prefixcommons/registry/{name}.jsonld"
    )
    return Converter.from_jsonld(url)


def get_monarch_converter() -> Converter:
    """Get the Prefix Commons-maintained Monarch context."""
    return get_prefixcommons_converter("monarch_context")


def get_go_converter() -> Converter:
    """Get the Prefix Commons-maintained GO context."""
    return get_prefixcommons_converter("go_context")


def get_bioregistry_converter(web: bool = False, **kwargs: Any) -> Converter:
    """Get the latest extended prefix map from the Bioregistry [hoyt2022]_.

    :param web: If false, tries to import :mod:`bioregistry` and use
        :func:`bioregistry.get_converter` to get the converter. Otherwise,
        falls back to using the GitHub-hosted EPM export.
    :param kwargs:
        Keyword arguments to pass to :meth`:curies.Converter.from_extended_prefix_map`
        when using web-based loading.
    :returns: A converter representing the Bioregistry, which includes
        a comprehensive collection of prefixes, prefix synonyms, and
        URI prefix synonyms.

    Short summary of the Bioregistry:

    1. It deduplicates and harmonizes dozens of different resources that
       curate partially overlapping and conflicting prefix maps
    2. It contains detailed CURIE prefix synonyms to support standardization
    3. It enforces the generation of a self-consistent extended prefix map

    The Bioregistry's primary prefixes are all standardized to be lowercase,
    have minimal punctuation, and be the most idiomatic possible. When this
    conflicts with your personal preferences/community preferences, you can
    chain another converter in front of the Bioregistry converter using
    :func:`curies.chain`.

    However, the Bioregistry itself presents a more
    sustainable way of documenting these deviations in a community-oriented way
    using its "context" configurations. See https://bioregistry.io/context/ for
    more information. One excellent example of a community context is for the
    OBO community (see https://bioregistry.io/context/obo), which prioritizes
    OBO capitalized prefixes and makes a few minor changes for backwards compatibility
    (e.g., renaming Orphanet).

    .. [hoyt2022] `Unifying the identification of biomedical entities with the
       Bioregistry <https://www.nature.com/articles/s41597-022-01807-3>`_
    """
    if not web:
        try:
            import bioregistry
        except ImportError:  # pragma: no cover
            pass
        else:
            return bioregistry.manager.get_converter()
    url = f"{BIOREGISTRY_CONTEXTS}/bioregistry.epm.json"
    return Converter.from_extended_prefix_map(url, **kwargs)
